/*
 * AVI/MOV/MP4 Container Writer
 *
 * Copyright (c) 2006 vixy project
 *
 * This file is part of VIXY FLV Converter.
 *
 * 'VIXY FLV Converter' is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * 'VIXY FLV Converter' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "avformat_writer.h"

#define MKTAG(a,b,c,d) (a | (b << 8) | (c << 16) | (d << 24))

void avfwriter_global_init()
{
	av_register_all();
}

AVFWRITER* avfwriter_init(const char* fn)
{
	AVFWRITER* p = (AVFWRITER*)malloc(sizeof(AVFWRITER));
	memset(p, 0, sizeof(*p));

    p->fmt = guess_format(NULL, fn, NULL);
    if (!p->fmt) 
	{
		avfwriter_release(p);
		return NULL;
	}	
	
	if (p->fmt->video_codec != CODEC_ID_MPEG4)
	{
		avfwriter_release(p);
		return NULL;
	}

	p->oc = av_alloc_format_context();
    if (!p->oc) 
	{
 		avfwriter_release(p);
		return NULL;
	}
	
	p->oc->oformat = p->fmt;

    if (url_fopen(&p->oc->pb, fn, URL_WRONLY) < 0) 
	{
 		avfwriter_release(p);
		return NULL;
    }
	
	return p;
}

int avfwriter_set_m4v_header(AVFWRITER* p, 
	int width, int height, int bitrate,
	const uint8* extradata, int extradatasize)
{

    AVCodecContext *c;
    AVStream *st;

    st = av_new_stream(p->oc, 0);
    if (!st) 
	{
		return -1;
    }

    c = st->codec;
    c->codec_id = CODEC_ID_MPEG4;
	c->codec_type = CODEC_TYPE_VIDEO;	
	
	if (strcmp(p->oc->oformat->name, "avi") == 0)
	{
		c->codec_tag = MKTAG('D','I','V','X');
//		c->codec_tag = MKTAG('M','P','4','S');
	}
	
    c->bit_rate = bitrate;
    c->width = width;
    c->height = height;
	
	c->time_base.den = 30; // for 30Hz
    c->time_base.num = 1;  // for 30Hz
	
	if (extradatasize > 0)
	{
		c->extradata_size = extradatasize;
		c->extradata = av_malloc(c->extradata_size);
		memcpy(c->extradata, extradata, c->extradata_size);
	}

    if (strcmp(p->oc->oformat->name, "mp4") == 0 || 
		strcmp(p->oc->oformat->name, "mov") == 0 || 
		strcmp(p->oc->oformat->name, "3gp") == 0)
	{
        c->flags |= CODEC_FLAG_GLOBAL_HEADER;
	}
		
	p->video_st = st;
	return 0;
}

int avfwriter_set_mp3_header(AVFWRITER* p, MP3HEADER* mp3header)
{
    AVCodecContext *c;
    AVStream *st;

    st = av_new_stream(p->oc, 1);
    if (!st) {
        fprintf(stderr, "Could not alloc stream\n");
        exit(1);
    }

    c = st->codec;
    c->codec_id = CODEC_ID_MP3;
    c->codec_type = CODEC_TYPE_AUDIO;

    c->bit_rate = mp3header->bit_rate;
    c->sample_rate = mp3header->sample_rate;
    c->channels = mp3header->channels;
	c->frame_size= mp3header->frame_size;

	p->audio_st = st;

	return 0;
}

int avfwriter_write_header(AVFWRITER* p)
{
    if (av_set_parameters(p->oc, NULL) < 0) 
	{
		return -1;
    }

    av_write_header(p->oc);

	p->write_header = 1;
	return 0;
}

static int write_frame(AVFormatContext *oc, int stream_idx, int keyframe, int pts, const uint8* buf, int size)
{
	AVPacket pkt;	
	av_init_packet(&pkt);
	pkt.flags |= keyframe ? PKT_FLAG_KEY : 0;
	pkt.stream_index = stream_idx;
	pkt.data= (uint8*)buf;
	pkt.size= size;
	pkt.pts = pts;
	return av_write_frame(oc, &pkt);
}

int avfwriter_write_m4v_frame(AVFWRITER* p, int keyframe, int pts, const uint8* buf, int size)
{
	return write_frame(p->oc, p->video_st->index, keyframe, pts, buf, size);
}

int avfwriter_write_mp3_frame(AVFWRITER* p, int pts, const uint8* buf, int size)
{
	return write_frame(p->oc, p->audio_st->index, 1, pts, buf, size);
}

int avfwriter_release(AVFWRITER* p)
{
	int i;
	
	if (p->oc)
	{
		if (p->write_header)
		{
			av_write_trailer(p->oc);
		}

		for (i = 0; i < p->oc->nb_streams; i++)
		{
//			av_freep(&p->oc->streams[i]->codec);
			av_freep(&p->oc->streams[i]);
		}

		url_fclose(&p->oc->pb);
		av_free(p->oc);
	}
	
	return 0;
}

const char* avfwriter_get_format_name(AVFWRITER* p)
{
	if (p->oc == NULL) return NULL;
	if (p->oc->oformat == NULL) return NULL;
	return p->oc->oformat->name;
}

