/*
 * FLV to MPEG4 converter
 *
 * Copyright (c) 2006 vixy project
 *
 * This file contains the code that based on FFmpeg (http://ffmpeg.mplayerhq.hu/)
 * See original copyright notice in /FFMPEG_CREDITS and /FFMPEG_IMPORTS
 *
 * This file is part of VIXY FLV Converter.
 *
 * 'VIXY FLV Converter' is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * 'VIXY FLV Converter' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "m4v.h"
#include "bitwriter.h"
#include "flv.h"
#include "fetch.h"
#include "avformat_writer.h"
#include "logodata.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////// 
typedef struct _CONVCTX
{
	int width;
	int height;
	
	int frame;
	int icounter;
	
#define CC_UNKNOWN		0	
#define CC_NOT_EXISTS	1
#define CC_EXISTS		2
	int is_audio;
	int is_video;
	int header_written;
	int header_search_counter;

	int first_m4v_header;

	M4V_VOL vol;

} CONVCTX;


typedef struct 
{
	uint8* out_buf;
	M4V_VOL vol;
	AVFWRITER* avf;

	CONVCTX	conv;


} CTX;

#define VOL_TIME_BITS		5

////////////////////////////////////////////////////////////////////////////////////////////////////////////////// 
static const uint8 ff_mpeg4_y_dc_scale_table[32]={
//  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31
    0, 8, 8, 8, 8,10,12,14,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,34,36,38,40,42,44,46
};

static const uint8 ff_mpeg4_c_dc_scale_table[32]={
//  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31
    0, 8, 8, 8, 8, 9, 9,10,10,11,11,12,12,13,13,14,14,15,15,16,16,17,17,18,18,19,20,21,22,23,24,25
};


static void copy_vol(PICTURE* flv_pic, M4V_VOL* vol)
{
	vol->width = flv_pic->width;
	vol->height = flv_pic->height;
	vol->time_bits = VOL_TIME_BITS; // 0-31
}

static void copy_vop(PICTURE* flv_pic, M4V_VOP* vop, CONVCTX* c)
{
	vop->qscale = flv_pic->qscale;
	
	vop->time = c->frame % 30;
	vop->icount = (c->icounter + 29) / 30;
	vop->intra_dc_threshold = 99;
	
	if (flv_pic->picture_type == FLV_I_TYPE)
	{
		vop->picture_type = M4V_I_TYPE;
	}
	else
	{
		vop->picture_type = M4V_P_TYPE;
		vop->f_code = 1;
	}
}

static void copy_microblock(MICROBLOCK* flv_mb, M4V_MICROBLOCK* m4v_mb)
{
	int i;
	
	m4v_mb->dquant = flv_mb->dquant;
	memcpy(m4v_mb->block, flv_mb->block, sizeof(m4v_mb->block)); // !!!!!!!
	m4v_mb->intra = flv_mb->intra;
	m4v_mb->skip = flv_mb->skip;
	m4v_mb->mv_type = flv_mb->mv_type;
	
	memcpy(m4v_mb->mv_x, flv_mb->mv_x, sizeof(m4v_mb->mv_x)); // !!!!!!
	memcpy(m4v_mb->mv_y, flv_mb->mv_y, sizeof(m4v_mb->mv_y)); // !!!!!!

	// dc rescale
	if (m4v_mb->intra)
	{
		for (i = 0; i < 4; i++)
		{
			m4v_mb->block[i].block[0] *= 8;
			m4v_mb->block[i].block[0] /= ff_mpeg4_y_dc_scale_table[m4v_mb->qscale];
		}

		for (i = 4; i < 6; i++)
		{
			m4v_mb->block[i].block[0] *= 8;
			m4v_mb->block[i].block[0] /= ff_mpeg4_c_dc_scale_table[m4v_mb->qscale];
		}
	}
}

static int write_m4v_header(AVFWRITER* p, CONVCTX* c, BR* br, BW* bw)
{
	PICTURE picture;
	M4V_VOP vop;

	memset(&picture, 0, sizeof(picture));
	memset(&vop, 0, sizeof(vop));
	
	if (decode_picture_header(br, &picture) < 0) return -1;
	copy_vol(&picture, &c->vol);

	m4v_encode_m4v_header(bw, &c->vol, 0);

	m4v_stuffing(bw);
	flash_bw(bw);
	
	c->width = picture.width;
	c->height = picture.height;
		
	alloc_dcpred(&c->vol.dcpred, (c->width+15) / 16, (c->height+15) / 16);

	return avfwriter_set_m4v_header(p, c->width, c->height, 200 * 1000, bw->buf, bw->pos);
}

static int write_pad_not_coded_frames(AVFWRITER* p, CONVCTX* c, BW* bw, uint32 time)
{
	// if any timecode padding is needed, then pad.
	while (c->frame * 1000 / 30 < time)
	{
		M4V_VOP vop;
		memset(&vop, 0, sizeof(vop));
		vop.picture_type = M4V_P_TYPE;
		vop.time = c->frame % 30;
		vop.icount = (c->icounter + 29) / 30;
		m4v_encode_vop_header(bw, &vop, VOL_TIME_BITS, 1);
		m4v_stuffing(bw);

		flash_bw(bw);
		
		// write frame
		if (avfwriter_write_m4v_frame(p, 
			0, 
			0,//c->frame, 
			bw->buf, 
			bw->pos) < 0)
		{
			return -1;
		}
	
		clear_bw(bw);
		
		c->frame++;
		c->icounter++;
	}

	return 0;
}

static int write_m4v_picture_frame(AVFWRITER* p, CONVCTX* c, BR* br, BW* bw, PICTURE* flvpic, uint32 time)
{
	MICROBLOCK mb;
	M4V_VOP vop;
	M4V_MICROBLOCK m4v_mb;
	int x, y;
	int mb_width = (flvpic->width + 15) / 16;
	int mb_height = (flvpic->height + 15) / 16;

	memset(&vop, 0, sizeof(vop));

	copy_vop(flvpic, &vop, c);
	m4v_encode_vop_header(bw, &vop, VOL_TIME_BITS, 0);
		
	// transcode flv to mpeg4
	for (y = 0; y < mb_height; y++)
	{
		for (x = 0; x < mb_width; x++)
		{
			memset(&mb, 0, sizeof(mb));
			memset(&m4v_mb, 0, sizeof(m4v_mb));
			
			if (vop.picture_type == M4V_I_TYPE)
			{
				mb.intra = 1;
				if (decode_I_mb(br, &mb, flvpic->escape_type, flvpic->qscale) < 0) return -1;
				m4v_mb.qscale = vop.qscale;
				copy_microblock(&mb, &m4v_mb);
				m4v_encode_I_dcpred(&m4v_mb, &c->vol.dcpred, x, y);
				m4v_encode_I_mb(bw, &m4v_mb);
			}
			else
			{
				if (decode_P_mb(br, &mb, flvpic->escape_type, flvpic->qscale) < 0) return -1;
				m4v_mb.qscale = vop.qscale;
				copy_microblock(&mb, &m4v_mb);
				m4v_encode_I_dcpred(&m4v_mb, &c->vol.dcpred, x, y);
				m4v_encode_P_mb(bw, &m4v_mb);
			}
		}
	}

	m4v_stuffing(bw);
	flash_bw(bw);
	
	// write frame
	if (avfwriter_write_m4v_frame(p, 
		vop.picture_type == M4V_I_TYPE, 
		0,//c->frame, 
		bw->buf, 
		bw->pos) < 0)
	{
		return -1;
	}

	c->frame++;
	c->icounter++;

	return 0;
}


static int write_m4v_frame(AVFWRITER* p, CONVCTX* c, BR* br, BW* bw, uint32 time)
{
	PICTURE picture;

	memset(&picture, 0, sizeof(picture));
	init_dcpred(&c->vol.dcpred);
	
	if (decode_picture_header(br, &picture) < 0) return -1;
	if (c->width != picture.width || c->height != picture.height) return -1; //size changed..
	
	copy_vol(&picture, &c->vol);

	if (picture.picture_type == FLV_I_TYPE)
	{
		c->icounter = 0;
		if (c->first_m4v_header)
		{
			m4v_encode_m4v_header(bw, &c->vol, time);
			c->first_m4v_header = 0;
		}
	}
	else
	{
		if (write_pad_not_coded_frames(p, c, bw, time) < 0) return -1;
	}

	if (write_m4v_picture_frame(p, c, br, bw, &picture, time) < 0)
	{
		return -1;
	}
	
	return 0;
}

static void write_mp3(CTX* p, const uint8* buf, uint32 size, uint32 time)
{
	avfwriter_write_mp3_frame(p->avf, 0, buf, size);
}

static int on_flv_packet(void* ptr, uint8 picture_type, const uint8* buf, uint32 size, uint32 time)
{
	CTX* p = (CTX*)ptr;

	switch (buf[0])
	{
	case 8: // Audio
		if (p->conv.is_audio == CC_EXISTS)
		{
			write_mp3(p, buf + 12, size - 16, time);
		}
		break;

	case 9: // Video
		if (p->conv.is_video == CC_EXISTS)
		{
			BR br;
			BW bw;
			init_br(&br, buf + 12, size - 16);
			init_bw(&bw, p->out_buf, PACKETBUFFER_SIZE);
			write_m4v_frame(p->avf, &p->conv, &br, &bw, time);
		}
		break;
	}
	
	return 0; // continueing loop
}

#define MKHEADER(a,b,c,d)	((((uint32)a)<<24) | (((uint32)b)<<16) | (((uint32)c)<<8) | (((uint32)d)<<0))

static int check_and_write_hader(CTX* p)
{
	if (p->conv.header_written == 0)
	{
		if (p->conv.is_video == CC_EXISTS && p->conv.is_audio == CC_EXISTS)
		{
			avfwriter_write_header(p->avf);
			p->conv.header_written = 1;
			return 0;
		}
		else if (p->conv.is_video == CC_EXISTS && 
				(p->conv.is_audio == CC_NOT_EXISTS || p->conv.is_audio == CC_UNKNOWN))
		{
			avfwriter_write_header(p->avf);
			p->conv.header_written = 1;
			return 0;
		}
	}
	
	return -1;
}

static int on_flv_packet_to_make_header(void* ptr, uint8 picture_type, const uint8* buf, uint32 size, uint32 time)
{
	CTX* p = (CTX*)ptr;

	switch (buf[0])
	{
	case 8: // Audio
		if (p->conv.is_audio == CC_UNKNOWN)
		{
			MP3HEADER h;
			uint32 header = MKHEADER(buf[12],buf[13],buf[14],buf[15]);
			if (decode_mp3_header(&h, header) == 0)
			{
				p->conv.is_audio = CC_EXISTS;
				avfwriter_set_mp3_header(p->avf, &h);
			}
			else
			{
				p->conv.is_audio = CC_NOT_EXISTS;
			}

			check_and_write_hader(p);
		}
		break;

	case 9: // Video
		if (p->conv.is_video == CC_UNKNOWN)
		{
			if (picture_type == 0x12) // I-frame
			{
				BR br;
				BW bw;
				init_br(&br, buf + 12, size - 16);
				init_bw(&bw, p->out_buf, PACKETBUFFER_SIZE);
				if (write_m4v_header(p->avf, &p->conv, &br, &bw) == 0)
				{
					p->conv.is_video = CC_EXISTS;
				}
				else
				{
					p->conv.is_video = CC_NOT_EXISTS;
				}
			}

			check_and_write_hader(p);
		}
		break;
	}
	
	if (p->conv.header_written)
	{
		return -1; // exit callback loop
	}
	
	p->conv.header_search_counter++;
	if (p->conv.header_search_counter > 30)
	{
		check_and_write_hader(p);
		return -1;
	}
	
	return 0; // continueing loop
}

static int write_logo_frames(AVFWRITER* p)
{
	uint8 tmp[256];
	BW bw;
	int frame = 0, icounter = 0;
	int i;

	init_bw(&bw, tmp, 256);

	// insert logo I-frame
	if (avfwriter_write_m4v_frame(p, 1, 0, vixy_logo_data, sizeof(vixy_logo_data)) < 0) return -1;

	// insert not_coded P-frames
	for (i = 0; i < 29; i++)
	{
		M4V_VOP vop;
		memset(&vop, 0, sizeof(vop));
		vop.picture_type = M4V_P_TYPE;
		vop.time = frame % 30;
		vop.icount = (icounter + 29) / 30;
		m4v_encode_vop_header(&bw, &vop, VOL_TIME_BITS, 1);
		m4v_stuffing(&bw);

		flash_bw(&bw);
		
		// write frame
		if (avfwriter_write_m4v_frame(p, 0, 0, bw.buf, bw.pos) < 0)
		{
			return -1;
		}
	
		clear_bw(&bw);
		
		frame++;
		icounter++;		
	}	
	
	return 0;
}



static void set_params_from_dst_format_name(CTX *ctx, const char* name)
{
	if (strcmp(name, "mp4") == 0)
	{
		ctx->conv.is_audio = CC_NOT_EXISTS; // skip audio
	}
	else if (strcmp(name, "avi") == 0)
	{
		ctx->conv.first_m4v_header = 1;
	}
	else if (strcmp(name, "mov") == 0)
	{
	}
}

int main (int argc, const char * argv[]) 
{
	CTX ctx;
	FILE* fp;
	FETCH* fetch;
	int r = -1;

	if (argc != 3)
	{
		fprintf(stderr, "usage: %s in.flv outfile\n", argv[0]);
		return -1;
	}

	avfwriter_global_init();

	fp = fopen(argv[1], "rb");
	
	memset(&ctx, 0, sizeof(ctx));
	ctx.out_buf = malloc(PACKETBUFFER_SIZE);
	memset(ctx.out_buf, 0, PACKETBUFFER_SIZE);
	memset(&ctx.vol, 0, sizeof(ctx.vol));
	
	ctx.avf = avfwriter_init(argv[2]);

	set_params_from_dst_format_name(&ctx, avfwriter_get_format_name(ctx.avf));
	
	fetch = FETCH_createInstance(on_flv_packet_to_make_header, &ctx);
	FETCH_read(fetch, fp);
	FETCH_release(fetch);
	
	if (ctx.conv.header_written)
	{
		fseek(fp, 0L, SEEK_SET);
		fetch = FETCH_createInstance(on_flv_packet, &ctx);
		FETCH_read(fetch, fp);
		FETCH_release(fetch);
		
		write_logo_frames(ctx.avf);
		r = 0;
	}
	
	avfwriter_release(ctx.avf);
	free_dcpred(&ctx.conv.vol.dcpred);
	
	fclose(fp);
	free(ctx.out_buf);

	return r;
}

